package com.pseudocode.photolib.metadata;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;

import javax.imageio.ImageIO;

import org.apache.commons.imaging.ImageInfo;
import org.apache.commons.imaging.ImageReadException;
import org.apache.commons.imaging.Sanselan;
import org.apache.commons.imaging.common.IImageMetadata;
import org.apache.commons.imaging.common.ImageMetadata;
import org.apache.commons.imaging.common.RationalNumber;
import org.apache.commons.imaging.formats.jpeg.JpegImageMetadata;
import org.apache.commons.imaging.formats.jpeg.JpegPhotoshopMetadata;
import org.apache.commons.imaging.formats.tiff.TiffDirectory;
import org.apache.commons.imaging.formats.tiff.TiffField;
import org.apache.commons.imaging.formats.tiff.TiffImageMetadata;
import org.apache.commons.imaging.formats.tiff.constants.ExifTagConstants;
import org.apache.commons.imaging.formats.tiff.constants.GpsTagConstants;
import org.apache.commons.imaging.formats.tiff.constants.TiffFieldTypeConstants;
import org.apache.commons.imaging.formats.tiff.constants.TiffTagConstants;
import org.apache.commons.imaging.formats.tiff.taginfos.TagInfo;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.adobe.xmp.XMPConst;
import com.adobe.xmp.XMPDateTime;
import com.adobe.xmp.XMPDateTimeFactory;
import com.adobe.xmp.XMPException;
import com.adobe.xmp.XMPMeta;
import com.adobe.xmp.XMPMetaFactory;
import com.adobe.xmp.XMPUtils;
import com.adobe.xmp.options.PropertyOptions;
import com.adobe.xmp.options.SerializeOptions;
import com.pseudocode.photolib.utils.GPSUtil;
import com.pseudocode.photolib.utils.RationalUtil;

public class XMPUtil {

	private static final Logger log = LoggerFactory.getLogger(XMPUtil.class);

	public static final String SIDECAR_EXT = ".xmp";
	public static final String NS_MSPhoto = "http://ns.microsoft.com/photo/1.0/";
	public static final String NS_LR = "http://ns.adobe.com/lightroom/1.0/";
	public static final String NS_PHOTOLIB = "http://ns.pseudocode.com/photolib/3.0/";
	public static final String PREFIX_PHOTOLIB = "photolib";
	public static final String TAG_FILE_HASH = "FileHash";
	public static final String TAG_FILE_FORMAT = "FileFormat";
	public static final String TAG_FILE_NAME = "FileName";
	public static final String TAG_FILE_SIZE = "FileSize";

	static {
		try {
			XMPMetaFactory.getSchemaRegistry().registerNamespace(NS_PHOTOLIB, PREFIX_PHOTOLIB);

		} catch (XMPException ex) {
			log.error("Error registering namespace prefix", ex);
		}
	}

	private static final String UTF16LE = "UTF-16LE";

	//StripOffsets, RowsPerStrip, StripByteCounts, JPEGInterchangeFormat, and JPEGInterchangeFormatLength
	private static final int TIFF_STRIP_OFFSETS = 0x0111;
	private static final int TIFF_ROWS_PER_STRIP = 0x0116;
	private static final int TIFF_STRIP_BYTE_COUNTS = 0x0117;
	private static final int TIFF_JPEG_INTERCHANGE_FORMAT = 0x0201;
	private static final int TIFF_JPEG_INTERCHANGE_FORMAT_LENGTH = 0x0202;
	private static final int TIFF_EXIF_PADDING = 0xEA1C;
	private static final int TIFF_XP_TITLE = 0x9C9B;
	private static final int TIFF_XP_COMMENT = 0x9C9C;
	private static final int TIFF_XP_AUTHOR = 0x9C9D;
	private static final int TIFF_XP_KEYWORDS = 0x9C9E;
	private static final int TIFF_XP_SUBJECT = 0x9C9F;
	private static final int EXIF_DATE_TIME = 0x0132;
	private static final int EXIF_DATE_TIME_ORIGINAL = 0x9003;
	private static final int EXIF_DATE_TIME_DIGITIZED = 0x9004;
	private static final int EXIF_SUB_SEC_TIME = 0x9290;
	private static final int EXIF_SUB_SEC_TIME_ORIGINAL = 0x9291;
	private static final int EXIF_SUB_SEC_TIME_DIGITIZED = 0x9292;
	private static final int GPS_GPS_LATITUDE = 0x0002;
	private static final int GPS_GPS_LONGITUDE = 0x0004;
	private static final int GPS_GPS_DEST_LATITUDE = 0x0014;
	private static final int GPS_GPS_DEST_LONGITUDE = 0x0016;
	private static final int GPS_GPS_LATITUDE_REF = 0x0001;
	private static final int GPS_GPS_LONGITUDE_REF = 0x0003;
	private static final int GPS_GPS_DEST_LATITUDE_REF = 0x0013;
	private static final int GPS_GPS_DEST_LONGITUDE_REF = 0x0015;
	private static final int GPS_GPS_DATE_STAMP = 0x001D;
	private static final int EXIF_CUSTOM_RENDERED = 0xA401;
	private static final int EXIF_EXPOSURE_MODE = 0xA402;
	private static final int EXIF_WHITE_BALANCE = 0xA403;
	private static final int EXIF_FOCAL_LENGTH_IN_35MM_FILM = 0xA406;
	private static final int EXIF_OWNER_NAME = 0xA430;
	private static final int EXIF_SERIAL_NUMBER = 0xA431;
	private static final int EXIF_LENS_INFO = 0xA432;
	private static final int EXIF_LENS_MODEL = 0xA434;
	private static final int EXIF_LENS_SERIAL_NUMBER = 0xA435;
	private static final int EXIF_SENSITIVITY_TYPE = 0x8830;
	private static final int EXIF_RECOMMENDED_EXPOSURE_INDEX = 0x8832;

	private static final int FIELD_TYPE_UNKNOWN = -1;
	private static final int FIELD_TYPE_BYTE = 1;
	private static final int FIELD_TYPE_ASCII = 2;
	private static final int FIELD_TYPE_SHORT = 3;
	private static final int FIELD_TYPE_LONG = 4;
	private static final int FIELD_TYPE_RATIONAL = 5;
	private static final int FIELD_TYPE_SBYTE = 6;
	private static final int FIELD_TYPE_UNDEFINED = 7;
	private static final int FIELD_TYPE_SSHORT = 8;
	private static final int FIELD_TYPE_SLONG = 9;
	private static final int FIELD_TYPE_SRATIONAL = 10;
	private static final int FIELD_TYPE_FLOAT = 11;
	private static final int FIELD_TYPE_DOUBLE = 12;

	{
		// these are needed to be true constant for switch block
		// verify that the external lib constants haven't changed
		if (FIELD_TYPE_UNKNOWN != TiffFieldTypeConstants.FIELD_TYPE_UNKNOWN.type) {
			throw new IllegalStateException("FIELD_TYPE_UNKNOWN constant changed");
		}
		if (FIELD_TYPE_BYTE != TiffFieldTypeConstants.FIELD_TYPE_BYTE.type) {
			throw new IllegalStateException("FIELD_TYPE_BYTE constant changed");
		}
		if (FIELD_TYPE_ASCII != TiffFieldTypeConstants.FIELD_TYPE_ASCII.type) {
			throw new IllegalStateException("FIELD_TYPE_ASCII constant changed");
		}
		if (FIELD_TYPE_SHORT != TiffFieldTypeConstants.FIELD_TYPE_SHORT.type) {
			throw new IllegalStateException("FIELD_TYPE_SHORT constant changed");
		}
		if (FIELD_TYPE_LONG != TiffFieldTypeConstants.FIELD_TYPE_LONG.type) {
			throw new IllegalStateException("FIELD_TYPE_LONG constant changed");
		}
		if (FIELD_TYPE_RATIONAL != TiffFieldTypeConstants.FIELD_TYPE_RATIONAL.type) {
			throw new IllegalStateException("FIELD_TYPE_RATIONAL constant changed");
		}
		if (FIELD_TYPE_SBYTE != TiffFieldTypeConstants.FIELD_TYPE_SBYTE.type) {
			throw new IllegalStateException("FIELD_TYPE_SBYTE constant changed");
		}
		if (FIELD_TYPE_UNDEFINED != TiffFieldTypeConstants.FIELD_TYPE_UNDEFINED.type) {
			throw new IllegalStateException("FIELD_TYPE_UNDEFINED constant changed");
		}
		if (FIELD_TYPE_SSHORT != TiffFieldTypeConstants.FIELD_TYPE_SSHORT.type) {
			throw new IllegalStateException("FIELD_TYPE_SSHORT constant changed");
		}
		if (FIELD_TYPE_SLONG != TiffFieldTypeConstants.FIELD_TYPE_SLONG.type) {
			throw new IllegalStateException("FIELD_TYPE_SLONG constant changed");
		}
		if (FIELD_TYPE_SRATIONAL != TiffFieldTypeConstants.FIELD_TYPE_SRATIONAL.type) {
			throw new IllegalStateException("FIELD_TYPE_SRATIONAL constant changed");
		}
		if (FIELD_TYPE_FLOAT != TiffFieldTypeConstants.FIELD_TYPE_FLOAT.type) {
			throw new IllegalStateException("FIELD_TYPE_FLOAT constant changed");
		}
		if (FIELD_TYPE_DOUBLE != TiffFieldTypeConstants.FIELD_TYPE_DOUBLE.type) {
			throw new IllegalStateException("FIELD_TYPE_DOUBLE constant changed");
		}
	}

	/**
	 * Deserializes XMP metadata from XML
	 * @param xmpFile XMP sidecar file
	 * @return
	 * @throws FileNotFoundException 
	 * @throws XMPException 
	 */
	public static XMPMeta readXMP(File xmpFile)
			throws FileNotFoundException, XMPException {

		if (xmpFile == null) {
			throw new NullPointerException("xmpFile");
		}

		InputStream stream = null;
		try {
			log.trace("reading XMP sidecar "+xmpFile.getPath());

			stream = new FileInputStream(xmpFile);
			return XMPMetaFactory.parse(stream);

		} finally {
			try {
				if (stream != null) {
					stream.close();
				}
			} catch (Exception ex) {}
		}
	}

	/**
	 * Serializes XMP metadata as canonical XML
	 * @param xmpFile XMP sidecar file
	 * @param xmp XMP metadata
	 * @throws FileNotFoundException
	 * @throws XMPException
	 */
	public static void writeXMP(File xmpFile, XMPMeta xmp)
			throws FileNotFoundException, XMPException {

		if (xmpFile == null) {
			throw new NullPointerException("xmpFile");
		}
		if (xmp == null) {
			throw new NullPointerException("xmp");
		}

		if (!xmpFile.getParentFile().exists()) {
			xmpFile.getParentFile().mkdirs();
		}

		OutputStream stream = new FileOutputStream(xmpFile);
		try {
			log.info("writing XMP sidecar "+xmpFile);

			SerializeOptions options = new SerializeOptions(SerializeOptions.ENCODE_UTF8);
			options.setBaseIndent(0);
			options.setIndent("\t");
			options.setNewline("\n");
			options.setUseCompactFormat(false);
			options.setUseCanonicalFormat(true);
			options.setOmitPacketWrapper(true);
			xmp.sort();
			XMPMetaFactory.serialize(xmp, stream, options);

		} finally {
			try {
				stream.flush();
				stream.close();
			} catch (Exception ex) {}
		}
	}

	/**
	 * Ensures XMP sidecar file exists, otherwise constructs from image and generates.
	 * @param imageFile original image file
	 * @param xmpFile XMP sidecar file
	 * @param fileHash SHA-1 hash of original image
	 * @return
	 */
	public static XMPMeta ensureXmp(File imageFile, File xmpFile, String fileHash) {
		XMPMeta xmp = null;
		boolean needWriteXMP = false;

		if (xmpFile.exists()) {
			try {
				// read the XMP sidecar XML
				xmp = XMPUtil.readXMP(xmpFile);

				if (xmp != null &&
					!xmp.doesPropertyExist(XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_HASH)) {

					XMPUtil.extractAppendImageInfo(xmp, imageFile, fileHash);

					// ensure orientation is sync'd with actual image EXIF
					int orientation = XMPUtil.extractEXIFOrientation(imageFile);
					if (orientation > 0 && orientation <= 8) {
						xmp.setProperty(XMPConst.NS_TIFF, "Orientation", orientation);
					} else {
						xmp.deleteProperty(XMPConst.NS_TIFF, "Orientation");
					}

					needWriteXMP = true;
				}

			} catch (Exception ex) {
				log.error("XMP read error: "+xmpFile, ex);
			}
		}

		if (xmp == null) {
			try {
				// generate XMP from embedded metadata
				xmp = XMPUtil.buildXMP(imageFile, fileHash);

				needWriteXMP = true;

			} catch (Exception ex) {
				log.error("XMP generation error: "+imageFile, ex);
			}
		}

		if (needWriteXMP) {
			try {
				// save results so will be easier next time
				XMPUtil.writeXMP(xmpFile, xmp);

			} catch (Exception ex) {
				log.error("XMP write error: "+xmpFile, ex);
			}
		}
		return xmp;
	}

	/**
	 * Attempts to construct complete XMP from an image metadata & XMP sidecar
	 * @param imageFile original image file
	 * @param fileHash SHA-1 hash of original image
	 * @return
	 * @throws FileNotFoundException
	 * @throws XMPException 
	 */
	public static XMPMeta buildXMP(File imageFile, String fileHash)
			throws FileNotFoundException {

		if (imageFile == null) {
			throw new NullPointerException("imageFile");
		}
		if (!imageFile.exists()) {
			throw new FileNotFoundException(imageFile.toString());
		}

		// must merge embedded as both may be incomplete
		// Metadata Working Group guidelines specify that EXIF has
		// precedence over XMP, so read XMP first and overwrite with EXIF

		XMPMeta xmp = XMPMetaFactory.create();
		try {
			// read the XMP packet embedded in original
			extractAppendXMP(xmp, imageFile);
		} catch (Exception ex) {
			log.error("XMP extract error: "+imageFile, ex);
		}

		try {
			// read the EXIF embedded in original, append to xmp
			extractAppendEXIF(xmp, imageFile);
		} catch (Exception ex) {
			log.error("EXIF extract error: "+imageFile, ex);
		}

		try {
			extractAppendImageInfo(xmp, imageFile, fileHash);
		} catch (Exception ex) {
			log.error("XMP extract error: "+imageFile, ex);
		}

		return xmp;
	}

	/**
	 * Attempts to extract just the orientation field from embedded EXIF metadata
	 * @param imageFile original image file
	 * @return orientation value
	 */
	public static int extractEXIFOrientation(File imageFile) {
		final int INVALID = 0;

		try {
			IImageMetadata imageMetadata = Sanselan.getMetadata(imageFile);
			if (imageMetadata == null) {
				return INVALID;
			}

			if (!(imageMetadata instanceof JpegImageMetadata)) {
				log.warn("unknown embeded metadata ("+imageMetadata.getClass()+") "+imageFile);
				return INVALID;
			}

			TiffImageMetadata exif = ((JpegImageMetadata)imageMetadata).getExif();
			if (exif == null) {
				return INVALID;
			}

			TiffField field = exif.findField(TiffTagConstants.TIFF_TAG_ORIENTATION);
			return (field == null) ? INVALID : field.getIntValue();

		} catch (Exception ex) {
			log.error("EXIF read error: "+imageFile, ex);
			return INVALID;
		}
	}

	/**
	 * Attempts to extract just the orientation field from embedded EXIF metadata
	 * @param imageStream raw byte stream
	 * @param filename name of file stream
	 * @return orientation value
	 */
	public static int extractEXIFOrientation(InputStream imageStream, String filename) {
		final int INVALID = 0;

		try {
			IImageMetadata imageMetadata = Sanselan.getMetadata(imageStream, filename);
			if (imageMetadata == null) {
				return INVALID;
			}

			if (!(imageMetadata instanceof JpegImageMetadata)) {
				log.warn("unknown embeded metadata ("+imageMetadata.getClass()+") "+filename);
				return INVALID;
			}

			TiffImageMetadata exif = ((JpegImageMetadata)imageMetadata).getExif();
			if (exif == null) {
				return INVALID;
			}

			TiffField field = exif.findField(TiffTagConstants.TIFF_TAG_ORIENTATION);
			return (field == null) ? INVALID : field.getIntValue();

		} catch (Exception ex) {
			log.error("EXIF read error: "+filename, ex);
			return INVALID;
		}
	}

	/**
	 * Attempts to extract just the orientation field from embedded EXIF metadata
	 * @param imageData raw image bytes
	 * @return orientation value
	 */
	public static int extractEXIFOrientation(byte[] imageData) {
		final int INVALID = 0;

		try {
			IImageMetadata imageMetadata = Sanselan.getMetadata(imageData);
			if (imageMetadata == null) {
				return INVALID;
			}

			if (!(imageMetadata instanceof JpegImageMetadata)) {
				log.warn("unknown embeded metadata ("+imageMetadata.getClass()+") in raw bytes");
				return INVALID;
			}

			TiffImageMetadata exif = ((JpegImageMetadata)imageMetadata).getExif();
			if (exif == null) {
				return INVALID;
			}

			TiffField field = exif.findField(TiffTagConstants.TIFF_TAG_ORIENTATION);
			return (field == null) ? INVALID : field.getIntValue();

		} catch (Exception ex) {
			log.error("EXIF read error from raw bytes", ex);
			return INVALID;
		}
	}

	/**
	 * Attempts to extract image file data and append to XMP object
	 * @param xmp XMP metadata
	 * @param imageFile original image file
	 * @param fileHash SHA-1 hash of original image
	 * @throws XMPException
	 */
	private static void extractAppendImageInfo(XMPMeta xmp, File imageFile, String fileHash)
			throws XMPException {

		// ensure file-based properties exist and are correct (file name will be first if multiple copies exist)
		xmp.setProperty(XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_NAME, imageFile.getName());
		xmp.setProperty(XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_SIZE, imageFile.length());
		xmp.setProperty(XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_FORMAT, getExtension(imageFile.getName()));
		xmp.setProperty(XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_HASH, fileHash);

		try {
			ImageInfo info = Sanselan.getImageInfo(imageFile);

			xmp.setProperty(XMPConst.NS_EXIF, "ExifImageWidth", info.getWidth());
			xmp.setProperty(XMPConst.NS_EXIF, "ExifImageLength", info.getHeight());

		} catch (Exception ex) {
			log.error("ImageInfo error: "+imageFile.getPath(), ex);

			try {
				// load source image data, this is more expensive so only do in error situations
				BufferedImage image = ImageIO.read(imageFile);
				if (image != null) {
					xmp.setProperty(XMPConst.NS_EXIF, "ExifImageWidth", image.getWidth());
					xmp.setProperty(XMPConst.NS_EXIF, "ExifImageLength", image.getHeight());
					image.flush();

				} else {
					log.error("Image could not be read: "+imageFile.getPath());
				}
			} catch (Exception ex2) {
				log.error("Image could not be read: "+imageFile.getPath(), ex2);
			}
		}
	}

	/**
	 * Attempts to extract embedded XMP metadata and append to XMP object
	 * @param xmp XMP metadata
	 * @param imageFile original image file
	 * @throws ImageReadException
	 * @throws IOException
	 * @throws XMPException
	 */
	private static void extractAppendXMP(XMPMeta xmp, File imageFile)
			throws ImageReadException, IOException, XMPException {

		String xml = Sanselan.getXmpXml(imageFile);
		if (xml == null) {
			return;
		}

		log.trace("reading embeded XMP "+imageFile.getPath());
		XMPMeta metaExtract = XMPMetaFactory.parseFromString(xml);

		XMPUtils.appendProperties(metaExtract, xmp, true, true, true);
	}

	/**
	 * Attempts to extract embedded EXIF metadata and append to XMP object
	 * @param xmp
	 * @param imageFile
	 * @throws ImageReadException
	 * @throws IOException
	 * @throws XMPException
	 */
	private static void extractAppendEXIF(XMPMeta xmp, File imageFile)
			throws ImageReadException, IOException, XMPException {

		IImageMetadata imageMetadata = Sanselan.getMetadata(imageFile);
		if (imageMetadata == null) {
			return;
		}
		if (!(imageMetadata instanceof JpegImageMetadata)) {
			log.warn("unknown embeded metadata ("+imageMetadata.getClass()+") "+imageFile);
			return;
		}

		log.trace("reading embeded EXIF "+imageFile);
		JpegImageMetadata metadata = (JpegImageMetadata)imageMetadata;

		JpegPhotoshopMetadata photoshop = metadata.getPhotoshop();
		if (photoshop != null) {
			String ns = XMPConst.NS_PHOTOSHOP;
			String dateCreated = null,
				timeCreated = null;
			String digitDate = null,
				digitTime = null;

			for (Object item : photoshop.getItems()) {
				if (!(item instanceof ImageMetadata.Item)) {
					if (item != null) {
						log.warn("EXIF item: "+item.getClass());
					}
					continue;
				}

				ImageMetadata.Item imgItem = (ImageMetadata.Item)item;
				String tagName = formatTagName(imgItem.getKeyword());
				String value = imgItem.getText();

				// merge DateCreated and TimeCreated
				if ("DateCreated".equals(tagName) || "TimeCreated".equals(tagName)) {
					if ("DateCreated".equals(tagName)) {
						dateCreated = value;
					} else {
						timeCreated = value;
					}

					if (dateCreated != null && timeCreated != null) {
						if (dateCreated.length() == 8 && timeCreated.length() >= 6) {
							int year = Integer.parseInt(dateCreated.substring(0, 4));
							int month = Integer.parseInt(dateCreated.substring(4, 6));
							int day = Integer.parseInt(dateCreated.substring(6, 8));
							int hour = Integer.parseInt(timeCreated.substring(0, 2));
							int minute = Integer.parseInt(timeCreated.substring(2, 4));
							int second = Integer.parseInt(timeCreated.substring(4, 6));
							String tz = timeCreated.substring(6);

							value = String.format(
								"%04d-%02d-%02dT%02d:%02d:%02d%s",
								year,
								month,
								day,
								hour,
								minute,
								second,
								tz);
							xmp.setProperty(ns, "DateCreated", value);
						}
						dateCreated = timeCreated = null;
					}
					continue;
				}

				// merge DigitalCreationDate and DigitalCreationTime
				if ("DigitalCreationDate".equals(tagName) || "DigitalCreationTime".equals(tagName)) {
					if ("DigitalCreationDate".equals(tagName)) {
						digitDate = value;
					} else {
						digitTime = value;
					}

					if (digitDate != null && digitTime != null) {
						if (digitDate.length() == 8 && digitTime.length() >= 6) {
							int year = Integer.parseInt(digitDate.substring(0, 4));
							int month = Integer.parseInt(digitDate.substring(4, 6));
							int day = Integer.parseInt(digitDate.substring(6, 8));
							int hour = Integer.parseInt(digitTime.substring(0, 2));
							int minute = Integer.parseInt(digitTime.substring(2, 4));
							int second = Integer.parseInt(digitTime.substring(4, 6));
							String tz = digitTime.substring(6);

							value = String.format(
								"%04d-%02d-%02dT%02d:%02d:%02d%s",
								year,
								month,
								day,
								hour,
								minute,
								second,
								tz);
							xmp.setProperty(ns, "DigitalCreationDate", value);
						}
						digitDate = digitTime = null;
					}
					continue;
				}

				try {
					if (xmp.doesPropertyExist(ns, tagName)) {
						xmp.deleteProperty(ns, tagName);
					}
					xmp.setProperty(ns, tagName, value);

					log.trace("PS value: {"+ns+"}:"+tagName+" => "+value);

				} catch (Exception ex) {
					log.error("PS error: {"+ns+"}:"+imgItem.toString(), ex);
				}
			}
		}

		TiffImageMetadata exif = metadata.getExif();
		if (exif != null) {
			for (Object dir : exif.getDirectories()) {
				if (!(dir instanceof TiffImageMetadata.Directory)) {
					if (dir != null) {
						log.warn("Unknown EXIF dir: "+dir.getClass());
					}
					continue;
				}
				TiffImageMetadata.Directory tiffDir = (TiffImageMetadata.Directory)dir;
				String ns = xmpNamespace(tiffDir);
				if (ns == null) {
					continue;
				}

				for (Object item : tiffDir.getItems()) {
					if (!(item instanceof TiffImageMetadata.Item)) {
						if (item != null) {
							log.warn("Unknown EXIF item: "+item.getClass());
						}
						continue;
					}
	
					TiffImageMetadata.Item tiffItem = (TiffImageMetadata.Item)item;
					TiffField field = tiffItem.getTiffField();
					String tagName = formatTagName(field.getTagName());

					try {
						if (xmp.doesPropertyExist(ns, tagName)) {
							xmp.deleteProperty(ns, tagName);
						}
						setProperty(xmp, ns, field, exif);
	
					} catch (Exception ex) {
						log.error("EXIF-to-XMP error: "+ns+':'+field.toString(), ex);
					}
				}
			}
		}
	}

	/**
	 * Helper for setting EXIF properties on XMP metadata
	 * @param xmp XMP metadata to append
	 * @param ns namespace
	 * @param field TIFF field
	 * @param exif EXIF metadata
	 * @throws ImageReadException
	 * @throws XMPException
	 */
	private static void setProperty(XMPMeta xmp, String ns, TiffField field, TiffImageMetadata exif)
			throws ImageReadException, XMPException {

		String tagName = formatTagName(field.getTagName());

		TagInfo tagInfo = field.tagInfo;
		Object value = tagInfo.getValue(field);

		if (value == null) {
			xmp.setProperty(ns, tagName, "");
			return;
		}

		log.trace("EXIF value: "+field.getFieldTypeName()+" => "+field.getTagName()+" => ("+value.getClass()+") "+value+"");

		if (field.directoryType == TiffDirectory.DIRECTORY_TYPE_SUB) {
			// as per Adobe (2010) XMP Specification, Part 2; description in Section 3.4.2, ignore these properties:
			switch (field.tag) {
				case TIFF_STRIP_OFFSETS:
				case TIFF_ROWS_PER_STRIP:
				case TIFF_STRIP_BYTE_COUNTS:
				case TIFF_JPEG_INTERCHANGE_FORMAT:
				case TIFF_JPEG_INTERCHANGE_FORMAT_LENGTH:
					// in XMP, these get dropped
					return;
			}

		} else if (field.directoryType == TiffDirectory.DIRECTORY_TYPE_EXIF) {
			switch (field.tag) {
				case TIFF_EXIF_PADDING:
				case EXIF_SUB_SEC_TIME:
				case EXIF_SUB_SEC_TIME_DIGITIZED:
				case EXIF_SUB_SEC_TIME_ORIGINAL:
					// in XMP, these get removed or merged into other values
					return;

				case EXIF_OWNER_NAME:
					// unknown tag
					tagName = "OwnerName";
					break;
				case EXIF_SERIAL_NUMBER:
					// unknown tag
					tagName = "SerialNumber";
					break;
				case EXIF_LENS_INFO:
					// unknown tag
					tagName = "LensInfo";
					break;
				case EXIF_LENS_MODEL:
					// unknown tag
					tagName = "LensModel";
					break;
				case EXIF_SENSITIVITY_TYPE:
					// unknown tag
					tagName = "SensitivityType";
					break;
				case EXIF_LENS_SERIAL_NUMBER:
					// unknown tag
					tagName = "LensSerialNumber";
					break;
				case EXIF_RECOMMENDED_EXPOSURE_INDEX:
					// unknown tag
					tagName = "RecommendedExposureIndex";
					break;
			}

		} else if (field.directoryType == TiffDirectory.DIRECTORY_TYPE_GPS) {

			switch (field.tag) {
				case GPS_GPS_LATITUDE_REF:
				case GPS_GPS_LONGITUDE_REF:
				case GPS_GPS_DEST_LATITUDE_REF:
				case GPS_GPS_DEST_LONGITUDE_REF:
				case GPS_GPS_DATE_STAMP:
					// in XMP, these get merged into other values
					return;
			}

			// this logic is based on Adobe (2010) XMP Specification, Part 3;
			// Tables 35 & 41; and description in Section 4.2.3.1

			if (value instanceof RationalNumber[]) {
				RationalNumber[] rationals = (RationalNumber[])value;
				if (rationals.length == 3) {
					TagInfo refTag = null;
					switch (tagInfo.tag) {
						case GPS_GPS_LATITUDE:
							refTag = GpsTagConstants.GPS_TAG_GPS_LATITUDE_REF;
							break;
						case GPS_GPS_LONGITUDE:
							refTag = GpsTagConstants.GPS_TAG_GPS_LONGITUDE_REF;
							break;
						case GPS_GPS_DEST_LATITUDE:
							refTag = GpsTagConstants.GPS_TAG_GPS_DEST_LATITUDE_REF;
							break;
						case GPS_GPS_DEST_LONGITUDE:
							refTag = GpsTagConstants.GPS_TAG_GPS_DEST_LONGITUDE_REF;
							break;
					}

					if (refTag != null) {
						TiffField refField = exif.findField(refTag);
						String ref = (refField == null) ? null : (String)refField.getValue();
						char direction = (ref == null || ref.trim().isEmpty()) ? ' ' : ref.trim().charAt(0);
						if (rationals[0].divisor == 1 &&
							rationals[1].divisor == 1 &&
							rationals[2].divisor == 1) {
							// XMP: if all 3 denominators are 1, then use integer version
							value = GPSUtil.toXMPString(rationals[0].numerator, rationals[1].numerator, rationals[2].numerator, direction);

						} else {
							// else combine into floating point and split back out
							double[] values = {
								RationalUtil.toDouble(rationals[0].numerator, rationals[0].divisor),
								RationalUtil.toDouble(rationals[1].numerator, rationals[1].divisor),
								RationalUtil.toDouble(rationals[2].numerator, rationals[2].divisor)
							};
							double coord = GPSUtil.toDegrees(values[0], values[1], values[2]);
							values = GPSUtil.toDegreesMinutesSeconds(coord);
							value = GPSUtil.toXMPString((int)values[0], (int)values[1], (int)values[2], direction);
						}
						xmp.setProperty(ns, tagName, value);
						return;

					} else if (GpsTagConstants.GPS_TAG_GPS_TIME_STAMP.equals(tagInfo)) {
						int hour = (rationals[0].divisor == 1) ? rationals[0].numerator : (int)RationalUtil.toDouble(rationals[0].numerator, rationals[0].divisor);
						int minute = (rationals[1].divisor == 1) ? rationals[1].numerator : (int)RationalUtil.toDouble(rationals[1].numerator, rationals[1].divisor);
						double second = (rationals[2].divisor == 1) ? (double)rationals[2].numerator : RationalUtil.toDouble(rationals[2].numerator, rationals[2].divisor);
						double nano = (second % 1.0) * 1e9;

						// As per Adobe (2010) XMP Specification, Part 2; description in Table 40
						// The GPSDateStamp tag is new in Exif 2.2. The GPS timestamp in Exif 2.1 does not include a date.
						// If not present, the date component for the XMP should be taken from exif:DateTimeOriginal, or if that
						// is also lacking from exif:DateTimeDigitized. If no date is available, do not write exif:GPSTimeStamp to XMP.
						TiffField dateStamp = exif.findField(GpsTagConstants.GPS_TAG_GPS_DATE_STAMP);
						if (dateStamp == null) {
							dateStamp = exif.findField(ExifTagConstants.EXIF_TAG_DATE_TIME_ORIGINAL);// DateTimeOriginal
							if (dateStamp == null) {
								dateStamp = exif.findField(ExifTagConstants.EXIF_TAG_CREATE_DATE);// DateTimeDigitized
							}
						}
						String dateStr = (dateStamp != null) ? dateStamp.getStringValue() : null;

						if (dateStr != null && dateStr.length() >= "yyyy:MM:dd".length()) {
							int year = Integer.parseInt(dateStr.substring(0, 4));
							int month = Integer.parseInt(dateStr.substring(5, 7));
							int day = Integer.parseInt(dateStr.substring(8, 10));

							XMPDateTime date = XMPDateTimeFactory.create(year, month, day, hour, minute, (int)second, (int)nano);
							// TODO: should this inspect other time stamp for cue?
							// iPhone 4/iOS 5 saves GPSTimeStamp as UTC
							xmp.setPropertyDate(ns, tagName, XMPDateTimeFactory.convertToLocalTime(date));
							// iPhone 3G saves as local time zone.
							//meta.setPropertyDate(ns, tagName, XMPDateTimeFactory.setLocalTimeZone(date));
						}
						return;
					}
				}
			}

		} else if (field.directoryType == TiffDirectory.DIRECTORY_TYPE_ROOT) {
			switch (field.tag) {
				case TIFF_EXIF_PADDING:
					// in XMP, these get dropped
					return;
				case TIFF_XP_TITLE:
				case TIFF_XP_COMMENT:
				case TIFF_XP_AUTHOR:
				case TIFF_XP_KEYWORDS:
				case TIFF_XP_SUBJECT:
					if (value instanceof byte[]) {
						try {
							// decode bytes as little endian UTF-16
							byte[] bytes = (byte[])value;
							if (bytes.length < 2) {
								return;
							}

							value = new String(bytes, 0, bytes.length-2, UTF16LE).trim();
							if (field.tag == TIFF_XP_KEYWORDS) {
								value = ((String)value).split("\\s*;\\s*");
							}

						} catch (UnsupportedEncodingException ex) {}
					}
				case EXIF_CUSTOM_RENDERED:
					ns = XMPConst.NS_EXIF;
					tagName = "CustomRendered";
					break;
				case EXIF_EXPOSURE_MODE:
					ns = XMPConst.NS_EXIF;
					tagName = "ExposureMode";
					break;
				case EXIF_WHITE_BALANCE:
					ns = XMPConst.NS_EXIF;
					tagName = "WhiteBalance";
					break;
				case EXIF_FOCAL_LENGTH_IN_35MM_FILM:
					ns = XMPConst.NS_EXIF;
					tagName = "FocalLengthIn35mmFilm";
					break;
			}

//		} else if (field.directoryType == TiffDirectory.DIRECTORY_TYPE_INTEROPERABILITY) {
//
//		} else {
//			System.out.println("***************** "+field.directoryType+"; "+field);
		}

		if (tagName.indexOf("UnknownTag") >= 0) {
			// http://www.sno.phy.queensu.ca/~phil/exiftool/TagNames/EXIF.html
			log.warn("Unknown tag in dir "+field.directoryType+": "+field);
		}

		// change the value type according to field
		switch (field.type) {
			case FIELD_TYPE_RATIONAL:
			case FIELD_TYPE_SRATIONAL:
				if (value.getClass().isArray()) {
					PropertyOptions arrayOptions = new PropertyOptions();
					arrayOptions.setArray(true);
					arrayOptions.setArrayOrdered(true);
	
					for (int i=0, count=Array.getLength(value); i<count; i++) {
						Object item = Array.get(value, i);
						if (item instanceof RationalNumber) {
							// format rational as integer or double
							RationalNumber rational = (RationalNumber)item;
							xmp.appendArrayItem(ns, tagName, arrayOptions, rational.numerator+"/"+rational.divisor, null);
						} else {
							xmp.appendArrayItem(ns, tagName, arrayOptions, item == null ? "" : item.toString(), null);
						}
					}
					return;
				}
				if (value instanceof RationalNumber) {
					// format rational as integer or double
					RationalNumber rational = (RationalNumber)value;
					xmp.setProperty(ns, tagName, rational.numerator+"/"+rational.divisor);
					return;
				}
				break;
			case FIELD_TYPE_ASCII:
				String altNS = null, altTagName = null;
				TiffField subSec = null;
				String dateStr;

				// Mappings as per Adobe (2010) XMP Specification, Part 3; Table 35
				switch (field.tag) {
					case EXIF_DATE_TIME:
						// tiff:DateTime => ModifyDate
						ns = XMPConst.NS_XMP;
						tagName = "ModifyDate";
						dateStr = field.getStringValue();
						subSec = exif.findField(ExifTagConstants.EXIF_TAG_SUB_SEC_TIME);
						break;
					case EXIF_DATE_TIME_ORIGINAL:
						// exif:DateTimeOriginal => photoshop:DateCreated
						altNS = XMPConst.NS_PHOTOSHOP;
						altTagName = "DateCreated";
						dateStr = field.getStringValue();
						subSec = exif.findField(ExifTagConstants.EXIF_TAG_SUB_SEC_TIME_ORIGINAL);
						break;
					case EXIF_DATE_TIME_DIGITIZED:
						// exif:DateTimeDigitized => xmp:CreateDate
						ns = XMPConst.NS_XMP;
						tagName = "CreateDate";
						dateStr = field.getStringValue();
						subSec = exif.findField(ExifTagConstants.EXIF_TAG_SUB_SEC_TIME_DIGITIZED);
						break;
					default:
						dateStr = null;
						break;
				}

				if (dateStr == null || dateStr.length() < "yyyy:MM:dd HH:mm:ss".length()) {
					break;
				}

				// convert EXIF date format "yyyy:MM:dd HH:mm:ss" to ISO-8601
				int year = Integer.parseInt(dateStr.substring(0, 4));
				int month = Integer.parseInt(dateStr.substring(5, 7));
				int day = Integer.parseInt(dateStr.substring(8, 10));
				int hour = Integer.parseInt(dateStr.substring(11, 13));
				int minute = Integer.parseInt(dateStr.substring(14, 16));
				int second = Integer.parseInt(dateStr.substring(17, 19));

				double nano = 0.0;
				if (subSec != null) {
					String subSecStr = subSec.getStringValue();
					if (subSecStr != null && !subSecStr.trim().isEmpty()) {
						// parse and convert to nanoseconds
						double val = Double.parseDouble(subSecStr.trim());
						if (val > 0.0) {
							nano = val * Math.pow(10, 9-subSecStr.trim().length());
						}
					}
				}

				XMPDateTime date = XMPDateTimeFactory.create(year, month, day, hour, minute, second, (int)nano);
				xmp.setPropertyDate(ns, tagName, XMPDateTimeFactory.setLocalTimeZone(date));
				if (altNS != null && altTagName != null) {
					xmp.setPropertyDate(altNS, altTagName, XMPDateTimeFactory.setLocalTimeZone(date));
				}
				return;

			default:
				// NOTE: the XP properties show up here too
				if (value instanceof byte[]) {
					xmp.setPropertyBase64(ns, tagName, (byte[])value);
					return;
				}
				break;
		}

		if (value.getClass().isArray()) {
			PropertyOptions arrayOptions = new PropertyOptions();
			arrayOptions.setArray(true);
			arrayOptions.setArrayOrdered(true);

			for (int i=0, count=Array.getLength(value); i<count; i++) {
				Object item = Array.get(value, i);
				xmp.appendArrayItem(ns, tagName, arrayOptions, item == null ? "" : item.toString(), null);
			}
			return;
		}

		// default to string
		if (value instanceof String) {
			xmp.setProperty(ns, tagName, ((String)value).trim());
		} else {
			xmp.setProperty(ns, tagName, value);
		}
	}

	/**
	 * Apache Sanselan to Adobe XMP namespace mapping
	 * @param tiffDir
	 * @return
	 */
	private static String xmpNamespace(TiffImageMetadata.Directory tiffDir) {
		switch (tiffDir.type) {
			case TiffDirectory.DIRECTORY_TYPE_EXIF:
			case TiffDirectory.DIRECTORY_TYPE_GPS:
				return XMPConst.NS_EXIF;

			case TiffDirectory.DIRECTORY_TYPE_ROOT:
			case TiffDirectory.DIRECTORY_TYPE_SUB:
			case TiffDirectory.DIRECTORY_TYPE_INTEROPERABILITY:
				return XMPConst.NS_TIFF;

			case TiffDirectory.DIRECTORY_TYPE_THUMBNAIL:
			case TiffDirectory.DIRECTORY_TYPE_UNKNOWN:
			default:
				log.warn("Unknown TIFF dir: ("+tiffDir.type+") "+tiffDir);
				return XMPConst.NS_TIFF;
		}
	}

	/**
	 * Hacky XML name safety cleanup
	 * @param name
	 * @return
	 */
	private static String formatTagName(String name) {
		return name.replaceAll("[^A-Za-z0-9_\\-]", "");
	}

	/**
	 * Returns the extension of the filename
	 * @param name
	 * @return
	 */
	private static String getExtension(String name) {
		int dot = name.lastIndexOf('.');
		if (dot < 0) {
			return "";
		}

		return name.substring(dot).toLowerCase();
	}
}
